﻿
using System;
using STMBench7.Core;
using System.Collections;
using STMBench7.Impl.Core;

/*
 * A Hashtable that is (nearly) safe for transactions. Warning: SEMISPACE
 * sometimes has issues with the multiuseword for stm + hashing.
 */
namespace STMBench7.Compatability
{
    //
    // HashtableOptNoBk
    // Optimized version of the Hashtable without buckets and initialization.
    // The first node of the linked list is stored directly to the container array.
    //

    //#define ENABLE_HASHTABLE_COUNTERS


        class GenomeHashtable
        {
            private const int DEFAULT_NUM_BUCKETS = 1023;    // The total number of buckets
            private const int DEFAULT_BUCKET_CAPACITY = 4;   // One bucket can have up to DEFAULT_BUCKET_CAPACITY - 1 entries
            private const int DEFAULT_RESEIZE_FACTR = 2;     // When resizing is necessry increment with 2                        


            [BartokSTMExtensions.NoLogTrans] //allocation that does not escape "this"
            public GenomeHashtable(int numBuckets)
            {
                Initialize(numBuckets);
            }


            [BartokSTMExtensions.NoLogTrans] //allocation that does not escape "this"
            public GenomeHashtable()
            {
                Initialize(DEFAULT_NUM_BUCKETS);
            }


            [BartokSTMExtensions.NoLogTrans] //allocation that does not escape "this"
            private void Initialize(int numBuckets)
            {
                _numBuckets = numBuckets;
                _container = new GenomeHashtable.Entry[_numBuckets];
                _numRehashes = 0;

                for (int i = 0; i < _container.Length; i++)
                {
                    _container[i] = null;
                }
            }

            //
            // This assumes that key does not exists in the hashtable.
            // Therefore we don't check if the key exists and if it exists to
            // update the value. Instead use the overloaded operator "[]".
            //
            public void Add(Object key, Object value)
            {
                uint hashCode = HashSdbm(key) % (uint)this._numBuckets;
                GenomeHashtable.Entry newEntry = new GenomeHashtable.Entry(key, value);

                if (_container[hashCode] == null)
                {
                    _container[hashCode] = newEntry;
                }
                else
                {
                    //
                    // Bucket exists.
                    //

                    GenomeHashtable.Entry entry = _container[hashCode];
                    GenomeHashtable.Entry last = null;
                    int numEntriesInBucket = 0;

                    do
                    {
                        if ( Benchmark.TransactionSafeEquals(entry.key, key) )
                        {
                            entry.value = value;
                            return;
                        }

                        last = entry;
                        entry = entry.next;
                        numEntriesInBucket++;
                    } while (entry != null);

                    //
                    // Entry with this key was not found. Let's add it.
                    //

                    last.next = newEntry;

                    if (numEntriesInBucket == DEFAULT_BUCKET_CAPACITY)
                    {
                        Rehash();
                    }
                }
            }

            /**
             *Ryan 
             */
            public void Remove(Object key)
            {
                uint hashCode = HashSdbm(key) % (uint)this._numBuckets;

                if (_container[hashCode] == null)
                {
                    return; //not in table
                }
                else
                {
                    //
                    // Bucket exists.
                    //

                    GenomeHashtable.Entry entry = _container[hashCode];
                    GenomeHashtable.Entry last = null;

                    while (entry != null)
                    {
                        if ( Benchmark.TransactionSafeEquals(entry.key, key) )
                        {

                            if (last != null)
                            {
                                last.next = entry.next;
                            }
                            else
                            {
                                _container[hashCode] = entry.next;
                            }

                            return;
                        }

                        last = entry;
                        entry = entry.next;
                    }
                    //not found...
                }
            }

            public bool Contains(Object key)
            {
                uint hashCode = HashSdbm(key) % (uint)this._numBuckets;
                GenomeHashtable.Entry entry = _container[hashCode];

                if (entry == null)
                {
                    return false;
                }
                else
                {
                    do
                    {
                        if ( Benchmark.TransactionSafeEquals(entry.key, key) )
                        {
                            return true;
                        }
                        else
                        {
                            entry = entry.next;
                        }
                    } while (entry != null);
                }

                return false;
            }


            public Object Get(Object key)
            {
                uint hashCode = HashSdbm(key) % (uint)this._numBuckets;
                GenomeHashtable.Entry entry = _container[hashCode];

                if (entry == null)
                {
                    return null;
                }
                else
                {
                    do
                    {
                        if (Benchmark.TransactionSafeEquals(entry.key, key))
                        {
                            return entry.value;
                        }
                        else
                        {
                            entry = entry.next;
                        }
                    } while (entry != null);
                }

                return null;
            }

            private void Rehash()
            {
                GenomeHashtable newHashtable = new GenomeHashtable(this._numBuckets * DEFAULT_RESEIZE_FACTR);

                for (int i = 0; i < this._numBuckets; i++)
                {
                    GenomeHashtable.Entry entry = this._container[i];

                    while (entry != null)
                    {
                        newHashtable.Add(entry.key, entry.value);
                        entry = entry.next;
                    }
                }

                this._container = newHashtable._container;
                this._numBuckets = newHashtable._numBuckets;
                this._numRehashes++;
            }



            static uint HashSdbm(Object key)
            {
                //if (!((key is DesignObjImpl)||(key is DocumentImpl))) Console.WriteLine(key.GetType());

                return (uint) key.GetHashCode();

                /*
                uint hash = 0;
                uint c = 0;

                for (int i = 0; i < key.Length; i++)
                {
                    c = (uint)key[i];
                    hash = c + (hash << 6) + (hash << 16) - hash;
                }

                return hash;*/
            }

            public int Count
            {
                get
                {
                    int count = 0;

                    for (int i = 0; i < _container.Length; i++)
                    {
                        GenomeHashtable.Entry entry = _container[i];

                        while (entry != null)
                        {
                            count++;
                            entry = entry.next;
                        }
                    }

                    return count;
                }
            }

            /**
             *Ryan
             */
            public IEnumerator GetEnumerator()
            {
                /*LinkedList2 ll = new LinkedList2();

                for (int i = 0; i < _container.Length; i++)
                {
                    GenomeHashtable.Entry entry = _container[i];

                    while (entry != null)
                    {
                        ll.Add(entry.key);
                        entry = entry.next;
                    }
                }

                return ll.GetEnumerator();*/
                
                return new GHEnumerator(this);
            }

            public class GHEnumerator : IEnumerator
            {
                GenomeHashtable hashtable;
                int bucket;
                GenomeHashtable.Entry entry;

                public GHEnumerator(GenomeHashtable ht)
                {
                    hashtable = ht;
                    Reset();
                }

                public object Current
                {
                    get { 
                        return entry==null ? null : entry.key;
                    }
                }

                /*
                 * Advance to next element, return true if there really was a next element
                 */
                public bool MoveNext()
                {
                    if (entry == null)
                    {
                        if (bucket == 0) //we're before first element
                            { /* fall through */ }
                        else //we're past last element
                            return false;
                    }
                    else
                    {
                        entry = entry.next;
                    }

                    while (bucket < hashtable._numBuckets && entry == null)
                    {
                        entry = hashtable._container[bucket];
                        bucket++;
                    }

                    return entry != null;
                }

                /*
                 * Position *before*, not at, the first element.
                 */
                public void Reset()
                {
                    bucket = 0;
                    entry = null;
                }
            }

            //[System.TransactionalMemory.AtomicNotSupported]
            public void DumpCounters()
            {
                Console.WriteLine(@"
Hashtable Counters:
   rehash: {0}", this._numRehashes);
            }

            private int _numBuckets;
            private GenomeHashtable.Entry[] _container;

            private int _numRehashes;

            class Entry
            {

                [BartokSTMExtensions.NoLogTrans] //allocation that does not escape "this"
                public Entry(Object key, object value)
                {
                    this.key = key;
                    this.value = value;
                    this.next = null;
                }

                public Object key;
                public object value;
                public Entry next;
            }
        }
}
