﻿using System;
 
 
using System.Text;

using STMBench7.Annotations;
using STMBench7.Backend;

namespace STMBench7.Core
{
	/**
	 * Used to create and destroy assembly elements while
	 * maintaining the consistency of the data structure
	 * and the indexes.
	 */
	public class AssemblyBuilder : DesignObjBuilder {

		private readonly IdPool baseAssemblyIdPool, complexAssemblyIdPool; 

		private readonly Index/*<IntIndexKey,BaseAssembly>*/ baseAssemblyIdIndex;
		private readonly Index/*<IntIndexKey,ComplexAssembly>*/ complexAssemblyIdIndex;
		
		public AssemblyBuilder(Index/*<IntIndexKey,BaseAssembly>*/ baseAssemblyIdIndex,
				Index/*<IntIndexKey,ComplexAssembly>*/ complexAssemblyIdIndex) {
			this.baseAssemblyIdIndex = baseAssemblyIdIndex;
			baseAssemblyIdPool = BackendFactory.instance.CreateIdPool(Parameters.MaxBaseAssemblies);
			
			this.complexAssemblyIdIndex = complexAssemblyIdIndex;
			complexAssemblyIdPool = BackendFactory.instance.CreateIdPool(Parameters.MaxComplexAssemblies);
		}

        public virtual Assembly CreateAndRegisterAssembly(Module module, ComplexAssembly superAssembly)
        {
			if( (superAssembly == null) || (superAssembly.GetLevel() > 2) )
				return CreateAndRegisterComplexAssembly(module, superAssembly);
			
			return CreateAndRegisterBaseAssembly(module, superAssembly);
		}

        public void UnregisterAndRecycleAssembly(Assembly assembly) {
                if(assembly is ComplexAssembly)
                        UnregisterAndRecycleComplexAssembly((ComplexAssembly)assembly);
                else
                        UnregisterAndRecycleBaseAssembly((BaseAssembly)assembly);
        }

        public virtual void UnregisterAndRecycleBaseAssembly(BaseAssembly baseAssembly)
        {
			int baseAssemblyId = baseAssembly.GetId();
			baseAssemblyIdIndex.Remove(new IntIndexKey(baseAssemblyId));
			
			baseAssembly.GetSuperAssembly().RemoveSubAssembly(baseAssembly);
			
			ImmutableCollection/*<CompositePart>*/ componentsSet = baseAssembly.GetComponents();
			foreach(CompositePart component in componentsSet)
				baseAssembly.RemoveComponent(component);

			baseAssembly.ClearPointers();
			baseAssemblyIdPool.PutUnusedId(baseAssemblyId);
		}

        public virtual void UnregisterAndRecycleComplexAssembly(ComplexAssembly complexAssembly)
        {
			short currentLevel = complexAssembly.GetLevel();
			ComplexAssembly superAssembly = complexAssembly.GetSuperAssembly();
			
			if(superAssembly == null) 
				throw new RuntimeError("ComplexAssemblyFactory: root Complex Assembly cannot be removed!");
			
			superAssembly.RemoveSubAssembly(complexAssembly);
			
			ImmutableCollection/*<Assembly>*/ subAssembliesSet = complexAssembly.GetSubAssemblies();
			foreach(Assembly assembly in subAssembliesSet) {
				if(currentLevel > 2) UnregisterAndRecycleComplexAssembly((ComplexAssembly)assembly);
				else UnregisterAndRecycleBaseAssembly((BaseAssembly)assembly);
			}
			
			int id = complexAssembly.GetId();
			complexAssemblyIdIndex.Remove(new IntIndexKey(id));

			complexAssembly.ClearPointers();
			complexAssemblyIdPool.PutUnusedId(id);
		}
		
		public static bool aaa = false;

        private BaseAssembly CreateAndRegisterBaseAssembly(Module module, ComplexAssembly superAssembly)
        {
			int date = CreateBuildDate(Parameters.MinAssmDate, Parameters.MaxAssmDate);
			
            //can't do io in transaction anyway
            //if(aaa) Console.Error.WriteLine(baseAssemblyIdPool.ToString());
			int assemblyId = baseAssemblyIdPool.GetId();
			BaseAssembly baseAssembly = 
				designObjFactory.CreateBaseAssembly(assemblyId, CreateType(), date, module, superAssembly);
			//if(aaa) Console.Error.WriteLine(date + " " + assemblyId);
			
			baseAssemblyIdIndex.Put(new IntIndexKey(assemblyId), baseAssembly);

			superAssembly.AddSubAssembly(baseAssembly);

			return baseAssembly;
		}

        private ComplexAssembly CreateAndRegisterComplexAssembly(Module module, ComplexAssembly superAssembly)
        {
			int id = complexAssemblyIdPool.GetId();
			int date = CreateBuildDate(Parameters.MinAssmDate, Parameters.MaxAssmDate);
			ComplexAssembly complexAssembly = 
				designObjFactory.CreateComplexAssembly(id, CreateType(), date, module, superAssembly);

			try {
			    for(int i = 0; i < Parameters.NumAssmPerAssm; i++) 
				    CreateAndRegisterAssembly(module, complexAssembly);
		    }
		    catch(OperationFailedException e) {
			    foreach(Assembly subAssembly in complexAssembly.GetSubAssemblies())
				    UnregisterAndRecycleAssembly(subAssembly);
			    complexAssemblyIdPool.PutUnusedId(id);
			    complexAssembly.ClearPointers();
			    throw e;
		    }

			complexAssemblyIdIndex.Put(new IntIndexKey(id), complexAssembly);
			
			if(superAssembly != null) superAssembly.AddSubAssembly(complexAssembly);
			
			return complexAssembly;
		}
	}

}
