﻿using System;
 
 
using System.Text;

using STMBench7.Annotations;
using STMBench7.Backend;

namespace STMBench7.Core
{
	/**
	 * Used to create and destroy composite parts while
	 * maintaining the consistency of the data structure
	 * and the indexes.
	 */
	[Immutable]
	public class CompositePartBuilder : DesignObjBuilder {

		private readonly IdPool idPool; 
		
		private readonly Index/*<IntIndexKey,CompositePart>*/ compositePartIdIndex;
		private readonly DocumentBuilder documentBuilder;
		private readonly AtomicPartBuilder atomicPartBuilder;
		
		public CompositePartBuilder(Index/*<IntIndexKey,CompositePart>*/ compositePartIdIndex,
				Index/*<StringIndexKey,Document>*/ documentTitleIndex,
				Index/*<IntIndexKey,AtomicPart>*/ partIdIndex,
				Index/*<IntIndexKey,LargeSet<AtomicPart>>*/ partBuildDateIndex) {
			
			this.compositePartIdIndex = compositePartIdIndex;
			documentBuilder = new DocumentBuilder(documentTitleIndex);
			atomicPartBuilder = new AtomicPartBuilder(partIdIndex, partBuildDateIndex);
			idPool = BackendFactory.instance.CreateIdPool(Parameters.MaxCompParts);
		}
		
		/**
	     * Creates a component (composite part) with the documentation and
	     * a graph of atomic parts and updates all the relevand indexes.
	     */
        public virtual CompositePart CreateAndRegisterCompositePart()
        {
			int id = idPool.GetId();
			String type = CreateType();
				
			int buildDate;
			if(ThreadRandom.NextInt(100) < Parameters.YoungCompFrac)
			    buildDate = CreateBuildDate(Parameters.MinYoungCompDate, Parameters.MaxYoungCompDate);
			else
			    buildDate = CreateBuildDate(Parameters.MinOldCompDate, Parameters.MaxOldCompDate);

            
			Document documentation = null;
            AtomicPart[] parts = new AtomicPart[Parameters.NumAtomicPerComp];

            try {
                    documentation = documentBuilder.CreateAndRegisterDocument(id);
                    CreateAtomicParts(parts);
            }
            catch(OperationFailedException e) {
                    if(documentation != null) documentBuilder.UnregisterAndRecycleDocument(documentation);
                    foreach(AtomicPart part in parts)
                            if(part != null) atomicPartBuilder.UnregisterAndRecycleAtomicPart(part);
                    idPool.PutUnusedId(id);
                    throw e;
            }

		    CreateConnections(parts);
		    CompositePart component = designObjFactory.CreateCompositePart(id, type, buildDate, documentation);
    	 	
            foreach(AtomicPart part in parts) component.AddPart(part);
		    
            compositePartIdIndex.Put(new IntIndexKey(id), component);

            return component;
		}

        public virtual void UnregisterAndRecycleCompositePart(CompositePart compositePart)
        {
			int id = compositePart.GetId();
			compositePartIdIndex.Remove(new IntIndexKey(id));
			
			documentBuilder.UnregisterAndRecycleDocument(compositePart.GetDocumentation());
			
			foreach(AtomicPart atomicPart in compositePart.GetParts())
				atomicPartBuilder.UnregisterAndRecycleAtomicPart(atomicPart);
			
			ImmutableCollection/*<BaseAssembly>*/ usedInList = compositePart.GetUsedIn();
			foreach(BaseAssembly ownerAssembly in usedInList) {
				while(ownerAssembly.RemoveComponent(compositePart));
			}
			
			compositePart.ClearPointers();
			idPool.PutUnusedId(id);
		}

		/**
		 * Create all atomic parts of a given composite part.
		 */
        private void CreateAtomicParts(AtomicPart[] parts) 
        {
	        for(int partNum = 0; partNum < Parameters.NumAtomicPerComp; partNum++) {
	            AtomicPart part = atomicPartBuilder.CreateAndRegisterAtomicPart();
	            parts[partNum] = part;
	        }
        }
		
		/**
		 * Create connections between the parts.
		 */
        private void CreateConnections(AtomicPart[] parts)
        {

	    	// First, make all atomic parts be connected in a ring
	    	// (so that the resulting graph is fully connected)
			for(int partNum = 0; partNum < Parameters.NumAtomicPerComp; partNum++) {
				int connectTo = (partNum + 1) % Parameters.NumAtomicPerComp;
				parts[partNum].ConnectTo(parts[connectTo], CreateType(), 
						ThreadRandom.NextInt(Parameters.XYRange) + 1);
			}
			
	    	// Then add other connections randomly, taking into account
	    	// the NumConnPerAtomic parameter. The procedure is non-deterministic
	    	// but it should eventually terminate.
			for(int partNum = 0; partNum < Parameters.NumAtomicPerComp; partNum++) {
				AtomicPart currentPart = parts[partNum];
				
			    while(currentPart.GetNumToConnections() < Parameters.NumConnPerAtomic) {
			    	int connectTo = ThreadRandom.NextInt(Parameters.NumAtomicPerComp);
			    	parts[partNum].ConnectTo(parts[connectTo], CreateType(), 
			    			ThreadRandom.NextInt(Parameters.XYRange) + 1);
			    }
			}
		}
	}

}
