﻿using System;
 
 
using System.Text;

using STMBench7.Annotations;
using STMBench7.Core;
using STMBench7.Backend;
using STMBench7.Compatability;

namespace STMBench7.Correctness.Invariants
{
	/**
	 * Test of invariants of indexes, and their consistency
	 * with the main data structure. Checks also invariants
	 * of the composite parts that are not reachable from
	 * base assemblies.
	 */
	[Immutable]
	[ThreadLocal]
	public class IndexTest : InvariantTest {

		public static void CheckInvariants(Setup setup, bool initial, TraversedObjects traversedObjects) {

			// Complex assemblies
			Index/*<IntIndexKey,ComplexAssembly>*/ complexAssemblyIdIndex = setup.GetComplexAssemblyIdIndex();
			foreach(ComplexAssembly traversedAssembly in traversedObjects.complexAssemblies) {
				int id = traversedAssembly.GetId();
				ComplexAssembly indexedAssembly = (ComplexAssembly)complexAssemblyIdIndex.Get(new IntIndexKey(id));
				CheckIndexValue("ComplexAssembly.id", traversedAssembly, indexedAssembly, id);
			}
			CheckAllTraversed(complexAssemblyIdIndex, traversedObjects.complexAssemblies, "ComplexAssembly.id");
			
			// Base assemblies
			Index/*<IntIndexKey,BaseAssembly>*/ baseAssemblyIdIndex = setup.GetBaseAssemblyIdIndex();
			foreach(BaseAssembly traversedAssembly in traversedObjects.baseAssemblies) {
				int id = traversedAssembly.GetId();
				BaseAssembly indexedAssembly = (BaseAssembly)baseAssemblyIdIndex.Get(new IntIndexKey(id));
				CheckIndexValue("BaseAssembly.id", traversedAssembly, indexedAssembly, id);
			}
			CheckAllTraversed(baseAssemblyIdIndex, traversedObjects.baseAssemblies, "BaseAssembly.id");	
			
			// Composite parts
			Index/*<IntIndexKey,CompositePart>*/ compositePartIdIndex = setup.GetCompositePartIdIndex();
			foreach(CompositePart traversedComponent in traversedObjects.components) {
				int id = traversedComponent.GetId();
				CompositePart indexedComponent = (CompositePart)compositePartIdIndex.Get(new IntIndexKey(id));
				CheckIndexValue("CompositePart.id", traversedComponent, indexedComponent, id);
			}
			
			// Check invariants for components disconnected from the data structure
			// (and add those components to the set of traversed objects)
			foreach(CompositePart indexedComponent in compositePartIdIndex)
				if(! traversedObjects.components.Contains(indexedComponent))
					CompositePartTest.CheckInvariants(indexedComponent, initial, null, traversedObjects);
			
			// Documents
		    Index/*<StringIndexKey,Document>*/ documentTitleIndex = setup.GetDocumentTitleIndex();
		    foreach(Document traversedDocument in traversedObjects.documents) {
		    	String title = traversedDocument.GetTitle();
		    	int id = traversedDocument.GetDocumentId();
		    	
		    	Document indexedDocument = (Document)documentTitleIndex.Get(new StringIndexKey(title));
		    	CheckIndexValue("Document.title", traversedDocument, indexedDocument, id);
		    }
		    CheckAllTraversed(documentTitleIndex, traversedObjects.documents, "Document.title");

		    // Atomic parts (id index)
			Index/*<IntIndexKey,AtomicPart>*/ atomicPartIdIndex = setup.GetAtomicPartIdIndex();
		    foreach(AtomicPart traversedPart in traversedObjects.atomicParts) {
		    	int id = traversedPart.GetId();
		    	
		    	AtomicPart indexedPart = (AtomicPart)atomicPartIdIndex.Get(new IntIndexKey(id));
		    	CheckIndexValue("AtomicPart.id", traversedPart, indexedPart, id);
		    }
		    CheckAllTraversed(atomicPartIdIndex, traversedObjects.atomicParts, "AtomicPart.id");
		    
		    // Atomic parts (buildDate index)
		    Index/*<IntIndexKey,LargeSet<AtomicPart>>*/ atomicPartBuildDateIndex = setup.GetAtomicPartBuildDateIndex();
		    foreach(AtomicPart traversedPart in traversedObjects.atomicParts) {
		    	int id = traversedPart.GetId();
		    	LargeSet/*<AtomicPart>*/ sameBuildDateParts = 
		    		(LargeSet)atomicPartBuildDateIndex.Get(new IntIndexKey(traversedPart.GetBuildDate()));
		    	if(sameBuildDateParts == null || !sameBuildDateParts.Contains(traversedPart))
		    		ReportError("AtomicPart.buildDate", "element with id = " + id + " not in the index");
		    }

            foreach(LargeSet/*<AtomicPart>*/ sameBuildDateParts in atomicPartBuildDateIndex)
		    	foreach(AtomicPart indexedPart in sameBuildDateParts)
		    		if(! traversedObjects.atomicParts.Contains(indexedPart))
		    			ReportError("AtomicPart.buildDate", "index contains too many elements");
		}
		
		private static void ReportError(String index, String message) {
			throw new RuntimeError("Index " + index + ": " + message);
		}
		
		private static void CheckIndexValue(String index, Object traversedObject, Object indexedObject, int elementId) {
	    	if(indexedObject == null)		
	    		ReportError(index, "element with id = " + elementId + 
	    				" in the data structure but not in the index");
	    	if(indexedObject != traversedObject)
	    		ReportError(index, "element with id = "	+ elementId +
	    				" is a different object in the index and in the data structure");
		}
		
		//private static void CheckAllTraversed(Index<Object,Object> index, HashtableSet<Object> traversedSet, String indexName) {
        //private static void CheckAllTraversed<A,B>(Index<A,B> index, HashtableSet<B> traversedSet, String indexName) {
		//    foreach(B indexedObject in index)
		//    	if(! traversedSet.Contains(indexedObject))
		//    		ReportError(indexName, "index contains too many elements");
		//}
        private static void CheckAllTraversed(Index index, HashtableSet traversedSet, String indexName)
        {
            foreach (Object indexedObject in index)
                if (!traversedSet.Contains(indexedObject))
                    ReportError(indexName, "index contains too many elements");
        }
	}

}
