﻿using System;
 
 
using System.Text;

using STMBench7.Annotations;
using STMBench7.Core;
using STMBench7.Backend;
using STMBench7.Compatability;
using STMBench7.Impl.Backend;

namespace STMBench7.Correctness.Opacity
{
	/**
	 * An operation that compares two benchmark data structures
	 * and throws a RuntimeError if they differ. Used to check
	 * whether a concurrent execution ensures opacity.
	 */
	[Immutable]
	[ThreadLocal]
	public class StructureComparisonOperation : Operation  {

		private readonly Setup setup1, setup2;
		
		public StructureComparisonOperation(Setup setup1, Setup setup2) {
			this.setup1 = setup1;
			this.setup2 = setup2;
		}
		
		public int PerformOperation()
        {
            CheckModulesEqual();
			CheckDesignLibrariesEqual();
 
			return 0;
		}
		
		private void CheckModulesEqual() {
			Module module1 = setup1.GetModule(), module2 = setup2.GetModule();
			CheckReferences(module1, module2, setup1, 0, setup2, 0, "module");
			CheckDesignObjEqual(module1, module2);

			int id1 = module1.GetId(), id2 = module2.GetId();
			Manual man1 = module1.GetManual(), man2 = module2.GetManual();
			CheckReferences(man1, man2, module1, id1, module2, id2, "manual");
			CheckManualsEqual(man1, man2);
			
			ComplexAssembly root1 = module1.GetDesignRoot(), root2 = module2.GetDesignRoot();
			CheckReferences(root1, root2, module1, id1, module2, id2, "designRoot");
			CheckAssembliesEqual(root1, root2);
		}
		
		private void CheckDesignObjEqual(DesignObj obj1, DesignObj obj2) {
			int id1 = obj1.GetId(), id2 = obj2.GetId();
			CheckEqual(id1, id2, obj1, id1, obj2, id2, "id");
			CheckEqual(obj1.GetType(), obj2.GetType(), obj1, id1, obj2, id2, "type");
			CheckEqual(obj1.GetBuildDate(), obj2.GetBuildDate(), obj1, id1, obj2, id2, "buildDate");
		}
		
		private void CheckManualsEqual(Manual man1, Manual man2) {
			int id1 = man1.GetId(), id2 = man2.GetId();
			CheckEqual(id1, id2, man1, id1, man2, id2, "id");
			CheckEqual(man1.GetTitle(), man2.GetTitle(), man1, id1, man2, id2, "title");
			CheckEqual(man1.GetText(), man2.GetText(), man1, id1, man2, id2, "text");
			CheckReferences(man1.GetModule(), man2.GetModule(), man1, id1, man2, id2, "module");
		}

		private void CheckAssembliesEqual(Assembly assembly1, Assembly assembly2) {
			CheckDesignObjEqual(assembly1, assembly2);
			int id1 = assembly1.GetId(), id2 = assembly2.GetId();
			CheckReferences(assembly1.GetModule(), assembly2.GetModule(),
					assembly1, id1, assembly2, id2, "module");
			CheckReferences(assembly1.GetSuperAssembly(), assembly2.GetSuperAssembly(),
					assembly1, id1, assembly2, id2, "superAssembly");
			
			if(! Benchmark.TransactionSafeEquals(assembly1.GetType(),assembly2.GetType()))
				ReportError(assembly1, id1, assembly2, id2, "objects of different classes");
			
			if(assembly1 is ComplexAssembly) 
				CheckComplexAssembliesEqual((ComplexAssembly)assembly1, (ComplexAssembly)assembly2);
			else
				CheckBaseAssembliesEqual((BaseAssembly)assembly1, (BaseAssembly)assembly2);
		}
		
		private void CheckComplexAssembliesEqual(ComplexAssembly assembly1,
				ComplexAssembly assembly2) {
			int id1 = assembly1.GetId(), id2 = assembly2.GetId();
			CheckEqual(assembly1.GetLevel(), assembly2.GetLevel(), 
					assembly1, id1, assembly2, id2, "level");
			
			ImmutableCollectionImpl/*<Assembly>*/ subAss1 = (ImmutableCollectionImpl)assembly1.GetSubAssemblies(),
                subAss2 = (ImmutableCollectionImpl)assembly2.GetSubAssemblies();
			CheckEqual(subAss1.Size(), subAss2.Size(),
					assembly1, id1, assembly2, id2, "subAssemblies.size()");
			
			Iterator/*<Assembly>*/ it1 = new Iterator(subAss1.GetEnumerator()), it2 = new Iterator(subAss2.GetEnumerator());
			while(it1.HasNext())  {
				Assembly child1 = (Assembly)it1.Next(), child2 = (Assembly)it2.Next();
				CheckReferences(child1, child2, assembly1, id1, assembly2, id2, "subAssemblies");
				CheckAssembliesEqual(child1, child2);
			}
		}

		private void CheckBaseAssembliesEqual(BaseAssembly assembly1, 
				BaseAssembly assembly2) {
			int id1 = assembly1.GetId(), id2 = assembly2.GetId();
            ImmutableCollectionImpl/*<CompositePart>*/ comp1 = (ImmutableCollectionImpl)assembly1.GetComponents(),
                comp2 = (ImmutableCollectionImpl)assembly2.GetComponents();
			CheckEqual(comp1.Size(), comp2.Size(), 
					assembly1, id1, assembly2, id2, "components.size()");
			
			Iterator/*<CompositePart>*/ it1 = new Iterator(comp1.GetEnumerator()), it2 = new Iterator(comp2.GetEnumerator());
			while(it1.HasNext()) {
				CompositePart component1 = (CompositePart)it1.Next(), component2 = (CompositePart)it2.Next();
				CheckReferences(component1, component2, 
						assembly1, id1, assembly2, id2, "components");
				CheckEqual(component1.GetId(), component2.GetId(),
						assembly1, id1, assembly2, id2, "components");
			}
		}
		
		private void CheckDesignLibrariesEqual() {
			Index/*<IntIndexKey,CompositePart>*/ library1 = setup1.GetCompositePartIdIndex(),
				library2 = setup2.GetCompositePartIdIndex();
			CheckReferences(library1, library2, setup1, 0, setup2, 0, "CompositePart.id index");

			Iterator/*<CompositePart>*/ it1 = new Iterator(library1.GetEnumerator()), it2 = new Iterator(library2.GetEnumerator());
			while(it1.HasNext()) {
                if (!it2.HasNext())
					ReportError(setup1, 0, setup2, 0, "CompositePart.id index size differences");
				
				CompositePart component1 = (CompositePart)it1.Next(), component2 = (CompositePart)it2.Next();
				CheckReferences(component1, component2, setup1, 0, setup2, 0, "CompositePart.id index");
				CheckCompositePartsEqual(component1, component2);
			}
		}
		
		private void CheckCompositePartsEqual(CompositePart component1,
				CompositePart component2) {
			CheckDesignObjEqual(component1, component2);
			
			int id1 = component1.GetId(), id2 = component2.GetId();
			Document doc1 = component1.GetDocumentation(), doc2 = component2.GetDocumentation();
			CheckReferences(doc1, doc2, component1, id1, component2, id2, "documentation");
			CheckDocumentsEqual(doc1, doc2);

            ImmutableCollectionImpl/*<BaseAssembly>*/ usedIn1 = (ImmutableCollectionImpl)component1.GetUsedIn(),
                usedIn2 = (ImmutableCollectionImpl)component2.GetUsedIn();
			CheckEqual(usedIn1.Size(), usedIn2.Size(),
					component1, id1, component2, id2, "usedIn.size()");
			
			Iterator/*<BaseAssembly>*/ it1 = new Iterator(usedIn1.GetEnumerator()), it2 = new Iterator(usedIn2.GetEnumerator());
			while(it1.HasNext()) {
				BaseAssembly assembly1 = (BaseAssembly)it1.Next(), assembly2 = (BaseAssembly)it2.Next();
				CheckReferences(assembly1, assembly2,
						component1, id1, component2, id2, "usedIn");
				CheckEqual(assembly1.GetId(), assembly2.GetId(),
						component1, id1, component2, id2, "usedIn");
			}
			
			AtomicPart root1 = component1.GetRootPart(), root2 = component2.GetRootPart();
			CheckReferences(root1, root2, component1, id1, component2, id2, "rootPart");
			CheckEqual(root1.GetId(), root2.GetId(), component1, id1, component2, id2, "rootPart");
			
			LargeSetImpl/*<AtomicPart>*/ parts1 = (LargeSetImpl)component1.GetParts(),
				parts2 = (LargeSetImpl)component2.GetParts();
			CheckReferences(parts1, parts2, component1, id1, component2, id2, "parts");
			CheckEqual(parts1.Size(), parts2.Size(), component1, id1, component2, id2, 
					"parts.size()");

			Iterator/*<AtomicPart>*/ partsIt1 = new Iterator(parts1.GetEnumerator()), partsIt2 = new Iterator(parts2.GetEnumerator());
			while(partsIt1.HasNext()) {
				AtomicPart part1 = (AtomicPart)partsIt1.Next(), part2 = (AtomicPart)partsIt2.Next();
				CheckReferences(part1, part2, component1, id1, component2, id2, "parts");
				CheckAtomicPartsEqual(part1, part2);
			}
		}
		
		private void CheckDocumentsEqual(Document doc1, Document doc2) {
			int id1 = doc1.GetDocumentId(), id2 = doc2.GetDocumentId();
			CheckEqual(id1, id2, doc1, id1, doc2, id2, "id");
			CheckEqual(doc1.GetTitle(), doc2.GetTitle(), doc1, id1, doc2, id2, "title");
			CheckEqual(doc1.GetText(), doc2.GetText(), doc1, id1, doc2, id2, "text");
			CheckReferences(doc1.GetCompositePart(), doc2.GetCompositePart(),
					doc1, id1, doc2, id2, "part");
		}
		
		private void CheckAtomicPartsEqual(AtomicPart part1, AtomicPart part2) {
			CheckDesignObjEqual(part1, part2);
			int id1 = part1.GetId(), id2 = part2.GetId();
			CheckEqual(part1.GetX(), part2.GetX(), part1, id1, part2, id2, "x");
			CheckEqual(part1.GetY(), part2.GetY(), part1, id1, part2, id2, "y");
			CheckReferences(part1.GetPartOf(), part2.GetPartOf(),
					part1, id1, part2, id2, "partOf");

            //cast to avoid draggin in atomic unfriendly verions of getenumerator
            ImmutableCollectionImpl/*<Connection>*/ to1 = (ImmutableCollectionImpl)part1.GetToConnections(),
                to2 = (ImmutableCollectionImpl)part2.GetToConnections();
			CheckEqual(to1.Size(), to2.Size(), part1, id1, part2, id2, "to.size()");
			Iterator/*<Connection>*/ toIt1 = new Iterator(to1.GetEnumerator()), toIt2 = new Iterator(to2.GetEnumerator());
			while(toIt1.HasNext()) {
				Connection conn1 = (Connection)toIt1.Next(), conn2 = (Connection)toIt2.Next();
				CheckReferences(conn1, conn2, part1, id1, part2, id2, "to");
				CheckConnectionsEqual(conn1, conn2);
			}
		}
		
		private void CheckConnectionsEqual(Connection conn1, Connection conn2) {
			CheckEqual(conn1.GetType(), conn2.GetType(), conn1, 0, conn2, 0, "type");
			CheckEqual(conn1.GetLength(), conn2.GetLength(), conn1, 0, conn2, 0, "length");
			
			AtomicPart src1 = conn1.GetSource(), src2 = conn2.GetSource();
			CheckReferences(src1, src2, conn1, 0, conn2, 0, "from");
			CheckEqual(src1.GetId(), src2.GetId(), conn1, 0, conn2, 0, "from");
			
			AtomicPart dest1 = conn1.GetDestination(), dest2 = conn2.GetDestination();
			CheckReferences(dest1, dest2, conn1, 0, conn2, 0, "to");
			CheckEqual(dest1.GetId(), dest2.GetId(), conn1, 0, conn2, 0, "to");
		}
		
		private void ReportError(Object obj1, int id1, Object obj2, int id2,
				String message) {

                    throw new Exception(message);        
            //BOGUS: for some reason the below doesn't like being in a tx

			/*throw new RuntimeError("Object " + obj1.ToString() + " with id = " + id1 +
					" != object " + obj2.ToString() + " with id = " + id2 +
					": " + message);*/
		}
		
		private void CheckEqual(int val1, int val2, 
				Object obj1, int id1, Object obj2, int id2, String attribute) {
			if(val1 != val2)
				ReportError(obj1, id1, obj2, id2, "attribute " + attribute + " --> " +
						val1 + " != " + val2);
		}

		private void CheckEqual(String val1, String val2, 
				Object obj1, int id1, Object obj2, int id2, String attribute) {
			if(! Benchmark.TransactionSafeEquals(val1,val2))
				ReportError(obj1, id1, obj2, id2, "attribute " + attribute + " --> " +
						val1 + " != " + val2);
		}
		
		private void CheckReferences(Object val1, Object val2,
				Object obj1, int id1, Object obj2, int id2, String attribute) {
            
			if(val1 != null && val1 == val2)
				ReportError(obj1, id1, obj2, id2, "attribute " + attribute + " --> " +
						"references equal");
			else if( (val1 == null && val2 != null) || (val1 != null && val2 == null) )
				ReportError(obj1, id1, obj2, id2, "attribute " + attribute + " --> " + 
						"one (and only one) reference is null");
		}

		public OperationId GetOperationId() {
			return null;
		}
	}

}
