﻿using System;
using System.Collections;
 
 
using System.Text;
using STMBench7.Core;
using STMBench7.Compatability;
using System.Threading;

namespace STMBench7.Locking
{
	/**
	 * Implementation of the OperationExecutor used by the medium-grained
	 * locking method. It implements externally most of the locking
	 * for each operation. The locks for each operation are predefined,
	 * except for some complex assembly locks that may be acquired
	 * by some complex assembly methods (see the MGLockingComplexAssemblyImpl
	 * class).
	 * 
	 * Locking order:
	 * <ol>
	 * <li> Global structure lock,
	 * <li> Manual lock,
	 * <li> Base assemblies lock,
	 * <li> Composite parts lock,
	 * <li> Documents lock,
	 * <li> Atomic parts lock,
	 * <li> Complex assemblies locks: starting from the highest level
	 * 		(Parameters.NumAssmLevels), down to the lowest level
	 * 		(BASE_ASSEMBLY_LEVEL + 1).
	 * </ol>
	 */
	public class MGLockingOperationExecutor : OperationExecutor {

		private static readonly int BASE_ASSEMBLY_LEVEL = 1;

		/**
		 * Global structure lock is acquired by every operation:
		 * <ol>
		 * <li> in write mode, if the operation modifies the data structure (i.e.,
		 * adds or removes some elements) or modifies one of the indexes; and
		 * <li> in read mode otherwise.
		 * </ol>
		 */
		private static readonly Lock globalStructureReadLock,
				globalStructureWriteLock;

		/**
		 * Per-assembly-level locks.
		 */
		private static readonly ReentrantReadWriteLock[] assemblyLocks;
		private static readonly Lock[] assemblyReadLocks, assemblyWriteLocks;

		/**
		 * Per-object-type locks.
		 */
		private static readonly Lock compositePartReadLock, compositePartWriteLock,
				atomicPartReadLock, atomicPartWriteLock, documentReadLock,
				documentWriteLock, manualReadLock, manualWriteLock;

		/**
		 * Used for generating timestamps for the replay log.
		 */
		private static readonly AtomicInteger globalCounter = new AtomicInteger();

		/**
		 * Determines which per-assembly-level locks were acquired by the
		 * MGLockingComplexAssemblyImpl class.
		 */
//behavior of Java's static is implied in C#?
		public /*static*/ class AssemblyLocksAcquired {
			public readonly bool[] isReadAcquired, isWriteAcquired;

			public AssemblyLocksAcquired() {
				isReadAcquired = new bool[Parameters.NumAssmLevels + 1];
				isWriteAcquired = new bool[Parameters.NumAssmLevels + 1];
				Clear();
			}

			public void Clear() {
				for (int level = 0; level <= Parameters.NumAssmLevels; level++) {
					isReadAcquired[level] = false;
					isWriteAcquired[level] = false;
				}
			}
		}


        static GenomeHashtable assemblyLocksAcquiredTable = new GenomeHashtable();

        static AssemblyLocksAcquired ThreadLocalAssemblyLocksAcquired
        {
            get
            {
                lock (assemblyLocksAcquiredTable)
                {
                    if (!assemblyLocksAcquiredTable.Contains(Thread.CurrentThread))
                    {
                        assemblyLocksAcquiredTable.Add(Thread.CurrentThread, new AssemblyLocksAcquired());
                    }
                    return (AssemblyLocksAcquired)assemblyLocksAcquiredTable.Get(Thread.CurrentThread);
                }
            }
        }



		static MGLockingOperationExecutor() {
			ReentrantReadWriteLock globalStructureLock = new ReentrantReadWriteLock(
					false);
			globalStructureReadLock = globalStructureLock.ReadLock();
			globalStructureWriteLock = globalStructureLock.WriteLock();

			assemblyLocks = new ReentrantReadWriteLock[Parameters.NumAssmLevels + 1];
			assemblyReadLocks = new Lock[Parameters.NumAssmLevels + 1];
			assemblyWriteLocks = new Lock[Parameters.NumAssmLevels + 1];
			for (int level = 1; level <= Parameters.NumAssmLevels; level++) {
				assemblyLocks[level] = new ReentrantReadWriteLock(false);
				assemblyReadLocks[level] = assemblyLocks[level].ReadLock();
				assemblyWriteLocks[level] = assemblyLocks[level].WriteLock();
			}
			ReentrantReadWriteLock compositePartLock = new ReentrantReadWriteLock(
					false);
			compositePartReadLock = compositePartLock.ReadLock();
			compositePartWriteLock = compositePartLock.WriteLock();

			ReentrantReadWriteLock atomicPartLock = new ReentrantReadWriteLock(
					false);
			atomicPartReadLock = atomicPartLock.ReadLock();
			atomicPartWriteLock = atomicPartLock.WriteLock();

			ReentrantReadWriteLock documentLock = new ReentrantReadWriteLock(false);
			documentReadLock = documentLock.ReadLock();
			documentWriteLock = documentLock.WriteLock();

			ReentrantReadWriteLock manualLock = new ReentrantReadWriteLock(false);
			manualReadLock = manualLock.ReadLock();
			manualWriteLock = manualLock.WriteLock();
		}

		/**
		 * Called by a MGLockingComplexAssemblyImpl method to 
		 * read-lock a given complex assembly level.
		 */
		public static void ReadLockAssemblyLevel(int level) {
            AssemblyLocksAcquired threadAssemblyLocksAcquired = ThreadLocalAssemblyLocksAcquired;
				
             if (threadAssemblyLocksAcquired.isReadAcquired[level]
					|| threadAssemblyLocksAcquired.isWriteAcquired[level])
				return;

			assemblyReadLocks[level].Lock();
			threadAssemblyLocksAcquired.isReadAcquired[level] = true;
		}

		/**
		 * Called by a MGLockingComplexAssemblyImpl method to 
		 * write-lock a given complex assembly level.
		 */
		public static void WriteLockAssemblyLevel(int level) {
            AssemblyLocksAcquired threadAssemblyLocksAcquired = ThreadLocalAssemblyLocksAcquired;
                
			if (threadAssemblyLocksAcquired.isWriteAcquired[level])
				return;

			assemblyWriteLocks[level].Lock();
			threadAssemblyLocksAcquired.isWriteAcquired[level] = true;
		}

		private readonly Operation op;
		private readonly ArrayList/*<Lock>*/ locksToAcquire;
		private int lastOperationTimestamp;

		public MGLockingOperationExecutor(Operation op) {
			this.op = op;

			// Decide which locks should be acquired when a given
			// operation is executed
            locksToAcquire = new ArrayList/*<Lock>*/();
			OperationId operationId = op.GetOperationId();
			if(operationId == null) return;
			OperationType operationType = operationId.GetType();

			// Structural modification operations: exclusive access
			if (operationType == OperationType.STRUCTURAL_MODIFICATION) {
				locksToAcquire.Add(globalStructureWriteLock);
				return;
			}
			
			// Other operations: assume the structure is not modified
			locksToAcquire.Add(globalStructureReadLock);

			// Handling of individual cases
            OperationId id = op.GetOperationId();
			//switch (op.GetOperationId()) {
			//case OperationId.T1:
            //case OperationId.T6:
            //case OperationId.Q7:
            //case OperationId.ST1:
            //case OperationId.ST9:
            //case OperationId.OP1:
            //case OperationId.OP2:
            //case OperationId.OP3:
            if(id==OperationId.T1 || id==OperationId.T6 || id==OperationId.Q7 || id==OperationId.ST1
                || id==OperationId.ST9 || id==OperationId.OP1 || id==OperationId.OP2 || id==OperationId.OP3)
				locksToAcquire.Add(atomicPartReadLock);
			//	break;
            //case OperationId.T2a:
            //case OperationId.T2b:
            //case OperationId.T2c:
            //case OperationId.T3a:
            //case OperationId.T3b:
            //case OperationId.T3c:
            //case OperationId.ST6:
            //case OperationId.ST10:
            //case OperationId.OP9:
            //case OperationId.OP10:
            //case OperationId.OP15:
            else if (id == OperationId.T2a || id == OperationId.T2b || id == OperationId.T2c || id == OperationId.T3a
                || id == OperationId.T3b || id == OperationId.T3c || id == OperationId.ST6 || id == OperationId.ST10
                || id == OperationId.OP9 || id == OperationId.OP10 || id == OperationId.OP15)
            {
                locksToAcquire.Add(atomicPartWriteLock);
                //	break;
            }

            //case OperationId.T4:
            //case OperationId.ST2:
            else if (id == OperationId.T4 || id == OperationId.ST2)
            {
                locksToAcquire.Add(documentReadLock);
            }
            //	break;
            //case OperationId.T5:
            //case OperationId.ST7:
            else if (id == OperationId.T5 || id == OperationId.ST7)
            {
                locksToAcquire.Add(documentWriteLock);
                //	break;
            }
            else if (id == OperationId.Q6)
            //case OperationId.Q6:
            {
                locksToAcquire.Add(assemblyReadLocks[BASE_ASSEMBLY_LEVEL]);
                locksToAcquire.Add(compositePartReadLock);
                // Pre-lock all assembly levels to avoid deadlocks
                for (int level = Parameters.NumAssmLevels; level > 1; level--)
                    locksToAcquire.Add(assemblyReadLocks[level]);
            }
            //	break;

            //case OperationId.ST3:
            else if (id == OperationId.ST3)
            {
                locksToAcquire.Add(assemblyReadLocks[BASE_ASSEMBLY_LEVEL]);
                // Pre-lock all assembly levels to avoid deadlocks
                for (int level = Parameters.NumAssmLevels; level > 1; level--)
                    locksToAcquire.Add(assemblyReadLocks[level]);
            }
            //	break;

            else if (id == OperationId.ST4)
            {
                //case OperationId.ST4:
                locksToAcquire.Add(assemblyReadLocks[BASE_ASSEMBLY_LEVEL]);
                locksToAcquire.Add(documentReadLock);
            }
            //	break;

            else if (id == OperationId.ST5)
            {
                //case OperationId.ST5:
                locksToAcquire.Add(assemblyReadLocks[BASE_ASSEMBLY_LEVEL]);
                locksToAcquire.Add(compositePartReadLock);
            }
            //	break;

            else if (id == OperationId.ST8)
            {
                //case OperationId.ST8:
                locksToAcquire.Add(assemblyWriteLocks[BASE_ASSEMBLY_LEVEL]);
                // Pre-lock all assembly levels to avoid deadlocks
                for (int level = Parameters.NumAssmLevels; level > 1; level--)
                    locksToAcquire.Add(assemblyWriteLocks[level]);
                //	break;
            }

            else if (id == OperationId.OP4 || id == OperationId.OP5)
            {
                //case OperationId.OP4:
                //case OperationId.OP5:
                locksToAcquire.Add(manualReadLock);
                //	break;
            }

            else if (id == OperationId.OP6)
            //case OperationId.OP6:
            {
                //break;
            }

            else if (id == OperationId.OP7)
            {
                //case OperationId.OP7:
                locksToAcquire.Add(assemblyReadLocks[BASE_ASSEMBLY_LEVEL]);
                //	break;
            }
            else if (id == OperationId.OP8)
            {
                //case OperationId.OP8:
                locksToAcquire.Add(compositePartReadLock);
                //break;
            }

            else if (id == OperationId.OP11)
            {

                //case OperationId.OP11:
                locksToAcquire.Add(manualWriteLock);
                //break;
            }

            else if (id == OperationId.OP12)
            {
                //case OperationId.OP12:
                //	break;
            }

            else if (id == OperationId.OP13)
            {
                //case OperationId.OP13:
                locksToAcquire.Add(assemblyWriteLocks[BASE_ASSEMBLY_LEVEL]);
                //break;
            }

            else if (id == OperationId.OP14)
            {

                //case OperationId.OP14:
                locksToAcquire.Add(compositePartWriteLock);
                //	break;
            }
            else
            {
                //default:
                throw new RuntimeError("Unknown operation: "
                        + op.GetOperationId().ToString());
            }
		}

		public int Execute() {
			try {

				foreach (Lock _lock in locksToAcquire) _lock.Lock();
				return op.PerformOperation();
			} 
			finally {
				if (Parameters.sequentialReplayEnabled)
					lastOperationTimestamp = globalCounter.GetAndIncrement();

                AssemblyLocksAcquired threadAssemblyLocksAcquired = ThreadLocalAssemblyLocksAcquired;

				for (int level = 1; level <= Parameters.NumAssmLevels; level++) {
					if (threadAssemblyLocksAcquired.isReadAcquired[level])
						assemblyReadLocks[level].Unlock();
					if (threadAssemblyLocksAcquired.isWriteAcquired[level])
						assemblyWriteLocks[level].Unlock();
				}
				threadAssemblyLocksAcquired.Clear();

				foreach (Lock _lock in locksToAcquire) _lock.Unlock();
			}
		}

		public int GetLastOperationTimestamp() {
			return lastOperationTimestamp;
		}
	}

}
